test_template_host: "{{template_host}}"
test_template_path: "{{template_path}}"
test_template_mtime: "{{template_mtime}}"
test_template_uid: "{{template_uid}}"
test_template_fullpath: "{{template_fullpath}}"
test_template_run_date: "{{template_run_date}}"
test_ansible_managed: "{{ansible_managed}}"
