# Nothing to see here, this file is just empty to support a imp.load_source
# without doing anything
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


class test:
    def __init__(self, *args, **kwargs):
        pass
